/* Renewed City Growth, a GameScript for OpenTTD - by kero
 * Licence: GNU General Public Licence, version 2
 */

require("version.nut");
require("town.nut");

/* Import SuperLib for GameScript */
import("util.superlib", "SuperLib", 36);
Log <- SuperLib.Log;
Helper <- SuperLib.Helper;

class MainClass extends GSController 
{
	towns = null;
	current_date = 0;
	current_month = 0;
	current_year = 0;

	constructor()
	{
		::TGR_Array_Table <- {};
		::TownSupplied_Table <- {};
		::TownStockPilesCat_Table <- {};
		::TownSuppliedCat_Table <- {};
		::TownGoalsCat_Table <- {};
		::TownMonitorCheck_Table <- {};
		::SavedDataExist <- false;
	}
}

function MainClass::Save()
{
	Log.Info("Saving data...", Log.LVL_INFO);
	local table = {};
	
	// Building the tables of town's data
	foreach (i, town in this.towns)
	{
		::TGR_Array_Table[town.id] <- town.TGR_Array;
		::TownSupplied_Table[town.id] <- town.TownSupplied;
		::TownStockPilesCat_Table[town.id] <- town.TownStockPilesCat;
		::TownSuppliedCat_Table[town.id] <- town.TownSuppliedCat;
		::TownGoalsCat_Table[town.id] <- town.TownGoalsCat;
		::TownMonitorCheck_Table[town.id] <- town.MonitorCheck;
	}

	// Building the general data table with individual tables
	table.TGR_Array_Table <- ::TGR_Array_Table;
	table.TownSupplied_Table <- ::TownSupplied_Table;
	table.TownStockPilesCat_Table <- ::TownStockPilesCat_Table;
	table.TownSuppliedCat_Table <- ::TownSuppliedCat_Table;
	table.TownGoalsCat_Table <- ::TownGoalsCat_Table;
	table.TownMonitorCheck_Table <- ::TownMonitorCheck_Table;
	
	return table;
}

function MainClass::Load(version, saved_data)
{
	Log.Info("Loading data...", Log.LVL_INFO);

	::TGR_Array_Table <- saved_data.TGR_Array_Table;
	::TownSupplied_Table <- saved_data.TownSupplied_Table;
	::TownStockPilesCat_Table <- saved_data.TownStockPilesCat_Table;
	::TownSuppliedCat_Table <- saved_data.TownSuppliedCat_Table;
	::TownGoalsCat_Table <- saved_data.TownGoalsCat_Table;
	::TownMonitorCheck_Table <- saved_data.TownMonitorCheck_Table;
	::SavedDataExist = true;
}

function MainClass::Start()
{
	// Initializing the script
	local start_tick = GSController.GetTick();
	GSGame.Pause();
	Log.Info("Script initialisation...", Log.LVL_INFO);
	this.Init();
	GSGame.Unpause();
	local setup_duration = GSController.GetTick() - start_tick;
	
	Log.Info("Game setup done.", Log.LVL_INFO);
	Log.Info("Setup took " + setup_duration + " ticks.", Log.LVL_DEBUG);
	Log.Info("Happy playing !", Log.LVL_INFO);
	
	// Wait for the game to start
	GSController.Sleep(1);
	
	// Main loop
	while (true) 
	{
		this.ManageTowns();
		this.Sleep(60);
	}
}

function MainClass::Init()
{
	// Game settings
	GSGameSettings.SetValue("economy.town_growth_rate", 2);
	GSGameSettings.SetValue("economy.fund_buildings", 0);
	
	// Set current date
	this.current_date = GSDate.GetCurrentDate();
	this.current_month = GSDate.GetMonth(this.current_date);
	this.current_year = GSDate.GetYear(this.current_date);
	
	// Clear all signs (texts under citynames) if this is a saved game.
	Helper.ClearAllSigns();
	
	// Initialize cargo lists and variables
	this.InitCargoLists();
	
	// Create the towns list
	Log.Info("Create town list ... (can take a while on large maps)", Log.LVL_INFO);
	this.towns = this.CreateTownList();
}

function MainClass::InitCargoLists()
{
	/* Helper function to get a list of currently used cargo labels and their ID. */
	local list = GSCargoList();
	for(local i = 0; i < 32; ++i)
	{
		Log.Info(i + " = " + GSCargo.GetCargoLabel(i), Log.LVL_DEBUG);
	}
	
	/* Setup Global Cargo Variables, using several cargo categories:
	 * Cat 1: PaxMail
	 * Cat 2: GFood
	 * Cat 3: GGoods
	 * Cat 4: RawInd
	 * Cat 5: TransInd
	 */
	local industry_setting = GSController.GetSetting("industry_NewGRF");
	switch (industry_setting)
	{
		case(1): //Base industry
			break;
		case(2): // FIRS
			::CargoListCat <- [[0,2],[11,13,29],[5,25,28],[1,3,4,6,7,8,10,12,14,15,16,17,22,26,27],[9,18,19,20,21,23,24,31]]
			::CargoMinPopDemand <- [0,500,1500,4000,8000];
			::CargoPermille <- [70,40,35,20,15];
			::CargoDecay <- [0.4,0.2,0.2,0.1,0.1];
			::CargoCatNum <- 5;
			::CargoCatStrings <- ["PaxMail","GFood","GGoods","RawInd","TransInd"];
			break;
		case(3): //ECS
			break;
	}

	// Building a general list of all cargos, ordered by categories.
	::CargoList <- [];
	for (local i = 0; i < ::CargoListCat.len(); i++)
	{
		for (local j = 0; j < ::CargoListCat[i].len(); j++)
		{
			::CargoList.append(CargoListCat[i][j]);
		}
	}
}

function MainClass::CreateTownList()
{
	local towns_list = GSTownList();
	local towns_array = [];
	
	/* Make a squirrel array of all towns.
	 * For each town, an instance of GoalTown is created
	 * to store the data related to that town
	 * Note: c'est un vrai array, composé des instances GoalTown
	 */
	foreach (t, _ in towns_list)
	{
		towns_array.append(GoalTown(t));
	}
	
	return towns_array;
}


// A CheckForNewTowns function which isn't used at the moment.
/*function MainClass::CheckForNewTowns()
{
	// Do our town-list have fewer towns than the game?
	local world_towns = GSTownList();
	if (this.towns.len() < world_towns.Count())
	{
		// Create a GSList with all towns that we have
		local have_towns = GSList();
		foreach (town in this.towns)
		{
			have_towns.AddItem(town.id, 0);
		}
		
		// Reduce the world list of towns with those that we already have
		world_towns.RemoveList(have_towns);

		// Add the remaining towns to our list
		// Also, keep a list of newly added towns
		local new_towns = [];
		foreach (town_id, _ in world_towns)
		{
			local town = GoalTown(town_id);
			this.towns.append(town);
			new_towns.append(town);
		}
	}
}*/


function MainClass::ManageTowns()
{
	// Run the daily functions
	local date = GSDate.GetCurrentDate();
	local diff_date = date - current_date;
	if (diff_date == 0) return;
	else
	{
		/*foreach (town in this.towns)
		{
			town.ExpandTown();
		}*/
		
		current_date = date;
	}
	
	// Run the monthly functions
	local month = GSDate.GetMonth(date);
	local diff_month = month - current_month;
	if ( diff_month == 0) return;
	else
	{
		//GSGame.Pause();
		local month_tick = GSController.GetTick();
		Log.Info("Starting Monthly Updates...", Log.LVL_INFO);
				
		foreach (town in this.towns)
		{
			town.MonthlyManageTown();
		}
		
		current_month = month;
		local month_tick_duration = GSController.GetTick() - month_tick;
		Log.Info("Monthly Update took "+month_tick_duration+"ticks", Log.LVL_DEBUG);
		//GSGame.Unpause();
	}
	
	// Run the yearly functions - Nothing to do for now
	/*local year = GSDate.GetYear(date);
	local diff_year = year - current_year;
	if ( diff_year == 0) return;
	else
	{
		GSLog.Info("Starting Yearly Updates");
		current_year = year
	}*/
}
