/* Renewed City Growth, a GameScript for OpenTTD - by kero
 * Licence: GNU General Public Licence, version 2
 */


/*
 * For each town in the game, an instance of this class is created.
 * It holds the data related to a specific town.
 */
class GoalTown
{
	id = null;
	label_sign_id = -1;
	TGR_Array = [];		// Town growth rate array
	TGR_Average = null;	// Town growth rate average, calculated from the array
	TownSupplied = [];
	TownStockPilesCat = [];
	TownSuppliedCat = [];
	TownGoalsCat = [];
	MonitorCheck = false;
	
	constructor(town_id)
	{
		this.id = town_id;
		
		if (!::SavedDataExist)
		{
			TGR_Array = [0,0,0,0,0,0,0,0];
			GSTown.SetGrowthRate(this.id, GSTown.TOWN_GROWTH_NONE);
						
			// Setup all the town specific arrays
			local IndustrySetting = GSController.GetSetting("industry_NewGRF");
			switch(IndustrySetting)
			{
				case(1): // BASE
					break;
				case(2): // FIRS
					TownSupplied = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
					TownStockPilesCat = [0,0,0,0,0];
					TownSuppliedCat = [0,0,0,0,0];
					TownGoalsCat = [0,0,0,0,0];
					break;
				case(3): // ECS
					break;
			}
		}
		
		else
		{
			TGR_Array = ::TGR_Array_Table[this.id];
			TownSupplied = ::TownSupplied_Table[this.id];
			TownStockPilesCat = ::TownStockPilesCat_Table[this.id];
			TownSuppliedCat = ::TownSuppliedCat_Table[this.id];
			TownGoalsCat = ::TownGoalsCat_Table[this.id];
			// For whatever reason, MonitorCheck is loaded as int, not boolean, so we need to convert it back
			MonitorCheck = (TownMonitorCheck_Table[this.id] != 0);
			if (MonitorCheck)
			{
				this.UpdateTownText();
			}
		}
	}
	
	function MonthlyManageTown(); // <-- Main function to process town monthly
}

function GoalTown::MonthlyManageTown()
{
	local sumgoals = 0;
	local goaldiff = 0;
	local goaldiffpercent = 0.0;
	local curpop = GSTown.GetPopulation(this.id);
	local parsed_cat = 0;	// index of parsed category
	local new_town_growth_rate = null;
	// Ceci n'est pas bon. Il faut que ce soit adapté à chaque ensemble d'industries;
	TownSupplied = [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0];
	TownSuppliedCat = [0,0,0,0,0];
	TownGoalsCat = [0,0,0,0,0];
	
	// Difficulty factor:
	local d_percent = GSController.GetSetting("goal_scale_factor");
	local d_factor = d_percent > 0? d_percent / 100.0 : 1.0; // Use 1.0 if setting is missing

	// Checking cargo delivery
	for (local i = 0; i < ::CargoList.len(); i++)
	{
		for (local cid = GSCompany.COMPANY_FIRST; cid <= GSCompany.COMPANY_LAST; cid++)
		{
			if (GSCompany.ResolveCompanyID(cid) != GSCompany.COMPANY_INVALID)
			{
				TownSupplied[i] = GSCargoMonitor.GetTownDeliveryAmount(cid, ::CargoList[i], this.id, true);
				// Infos de contrôle
				if (TownSupplied[i] > 0)
				{
					Log.Info(GSTown.GetName(this.id)+" : "+GSCargo.GetCargoLabel(CargoList[i])+" "+TownSupplied[i], Log.LVL_DEBUG);
				}
			}
		}
		
		// Checking passengers delivery
		if (i == 0)
		{
			local paxcheck = TownSupplied[0];
			
			// If there is passenger delivery, we start the monitoring and all the calculations;
			if (paxcheck > 0)
			{
				this.MonitorCheck = true;
			}
			
			// If there is not pax delivery, decision depends whether city already was on monitoring:
			else
			{
				if (!this.MonitorCheck)
				{
					return;
				}
				
				else
				{
					// Checking whether there is passenger delivery
					for (local cid = GSCompany.COMPANY_FIRST; cid <= GSCompany.COMPANY_LAST; cid++)
					{
						if (GSCompany.ResolveCompanyID(cid) != GSCompany.COMPANY_INVALID)
						{
							paxcheck += GSCargoMonitor.GetTownPickupAmount(cid, ::CargoList[0], this.id, true);
						}
					}
					
					if (paxcheck == 0)
					{
						this.TGR_Array = [0,0,0,0,0,0,0,0];
						GSTown.SetGrowthRate(this.id, GSTown.TOWN_GROWTH_NONE);
						this.TownStockPilesCat = [0,0,0,0,0];
						this.MonitorCheck = false;
						this.StopMonitors();
						this.RemoveTownText();
						Log.Info("City of "+GSTown.GetName(this.id)+" is not monitored anymore.", Log.LVL_DEBUG);
						return;
					}
				}
			}
		}
				
		// Parsing et somme par catégorie des cargos livrés
		if (this.TownSupplied[i] > 0)			// On ne cherche que s'il y a une livraison sur ce cargo.
		{
			local is_found = false;
			while (parsed_cat < ::CargoCatNum)	// On parse les groupes un par un.
			{
				for (local k = 0; k < ::CargoListCat[parsed_cat].len(); k++)
				{
					if (::CargoList[i] == ::CargoListCat[parsed_cat][k])
					{
						this.TownSuppliedCat[parsed_cat] += this.TownSupplied[i];
						is_found = true;
						
						if ((k+1) == ::CargoListCat[parsed_cat].len())
						{
							parsed_cat++;
						}
						
						break;
					}
				}
				
				if (is_found == true)
				{
					break;
				}
				
				parsed_cat++;
			}
		}
	}
	
	// Calcul des goals
	for (local i = 0; i < CargoCatNum && curpop > ::CargoMinPopDemand[i]; i++)
	{
		this.TownGoalsCat[i] = max((((curpop  - ::CargoMinPopDemand[i]).tofloat() / 1000) * ::CargoPermille[i] * d_factor).tointeger(),1);
	}
	// Dans le cas où la pop d'une ville est trop faible pour donner un goal, on l'établit à 1.
	if (this.TownGoalsCat[0] < 1)
	{
		this.TownGoalsCat[0] = 1;
	}
	
	// Faire le calcul des sumgoals/diffgoal à la nouvelle sauce;
	for (local i = 0; i < CargoCatNum; i++)
	{
		// Stockage des biens cumulés avec le StockPiled et mise à jour du sumgoals
		if (this.TownGoalsCat[i] > 0)
		{
			sumgoals += this.TownGoalsCat[i];
			this.TownStockPilesCat[i] += this.TownSuppliedCat[i];
		}
		
		if (this.TownStockPilesCat[i] < this.TownGoalsCat[i])
		{
			goaldiff += (this.TownGoalsCat[i] - this.TownStockPilesCat[i]);
			this.TownStockPilesCat[i] = 0;
		}
		else 	// Si le StockPiled est supérieur on ampute ce qui est consommé
		{
			this.TownStockPilesCat[i] = ((this.TownStockPilesCat[i] - this.TownGoalsCat[i]) * (1 - ::CargoDecay[i])).tointeger();
			// Stocklimit = this.TownGoals[i] * 10;
			if (this.TownStockPilesCat[i] > 300 && this.TownStockPilesCat[i] > this.TownGoalsCat[i] * 10)
			{
				this.TownStockPilesCat[i] = this.TownGoalsCat[i] * 10;
			}
		}
		
		// Infos de contrôle
		Log.Info(GSTown.GetName(this.id)+" - Cat.: "+(i+1)+" : Stockpiled: "+this.TownStockPilesCat[i]+" Supplied: "+this.TownSuppliedCat[i]+" Goal: "+this.TownGoalsCat[i], Log.LVL_DEBUG);
	}

	// Convert the difference to a percent
	if ( sumgoals > 0 )
	{
		goaldiffpercent = (goaldiff.tofloat() / sumgoals.tofloat());
	}
	else
	{
		Log.Info("We have a sumgoals <= 0. How is that supposed to happen ?", Log.LVL_DEBUG);
	}
	// Infos de contrôle
	Log.Info("sumgoals="+sumgoals+" - goaldiff="+goaldiff+" - goaldiffpercent="+goaldiffpercent, Log.LVL_DEBUG);
	
	// Calculate the number of days till the next growth
	if (goaldiffpercent < 0.8)
	{
		local max_town_growth_rate = ((1 / (100 + curpop.tofloat())) * 50000);
		new_town_growth_rate = (max_town_growth_rate * (1+(10/(1-goaldiffpercent)-10))).tointeger();
		Log.Info("On calcule un TGR de : "+new_town_growth_rate, Log.LVL_DEBUG);
	}
	if (goaldiffpercent >= 0.8 || new_town_growth_rate > 2000)
	{
		new_town_growth_rate = 2000;
		Log.Info("On établit arbitrairement le new_TGR à"+new_town_growth_rate, Log.LVL_DEBUG);
	}
	
	local array_text_first = " ";
	foreach (i, val in this.TGR_Array)
	{
		array_text_first = array_text_first+"i"+i+"="+val+" ";
	}
	Log.Info(GSTown.GetName(this.id)+array_text_first+"Average="+this.TGR_Average, Log.LVL_DEBUG);
	
	// Introduire de la moyenne mobile dans le calcul du taux de croissance. 
	// Éventuellement dans un second temps,  pondérée/exponentielle ?
	local SumArray = 0;
	local i = 0;
	while (this.TGR_Array[i] > 0)
	{
		SumArray += this.TGR_Array[i];
		i++;
	}
	this.TGR_Array[i] = new_town_growth_rate;
	SumArray += this.TGR_Array[i];
	this.TGR_Average = (SumArray/(i+1)).tointeger();
	
	// Define the new growth rate using the average:
	GSTown.SetGrowthRate(this.id, this.TGR_Average);
	
	//Infos de contrôle
	local array_text = " ";
	foreach (i, val in this.TGR_Array)
	{
		array_text = array_text+"i"+i+"="+val+" ";
	}
	Log.Info(GSTown.GetName(this.id)+array_text+"Average="+this.TGR_Average, Log.LVL_DEBUG);

	// Décaler le contenu de l'array
	local j = 0;
	while (this.TGR_Array[7] > 0)
	{
		this.TGR_Array[j] = this.TGR_Array[j+1];
		j++;
		if (j == 7)
		{
			this.TGR_Array[j] = 0;
		}
	}
	
	this.UpdateTownText();
}

function GoalTown::UpdateTownText()
{
	// Add a sign by the town to display the current growth
	local sign_text = "Growth: "+GSTown.GetGrowthRate(this.id)+" days";
	if (this.label_sign_id == -1)
	{
		this.label_sign_id = GSSign.BuildSign(GSTown.GetLocation(this.id), sign_text);
	}
	else
	{
		if (GSSign.IsValidSign(this.label_sign_id))
		{
			GSSign.SetName(this.label_sign_id, sign_text);
		}
		else	
		{
			this.label_sign_id = GSSign.BuildSign(GSTown.GetLocation(this.id), sign_text);
		}
	}	

	// Town text will look like below:
	// Label:stored  supplied  required
	local text = "";
	for (local j = 0; j < ::CargoCatNum; j++)
	{
		text = text+::CargoCatStrings[j]+": "+this.TownStockPilesCat[j]+"/"+this.TownSuppliedCat[j]+"/"+this.TownGoalsCat[j]+" ";
	}
	
	// Sets the towns text box
	GSTown.SetText(this.id, text);
}

function GoalTown::RemoveTownText()
{
	local text = "No goals for now.";
	
	// Suppression de l'indication sur la map
	if (GSSign.IsValidSign(this.label_sign_id))
	{
		GSSign.RemoveSign(this.label_sign_id);
		this.label_sign_id = -1;
	}
	
	// Suppression du texte dans la townbox
	GSTown.SetText(this.id, text);
}

function GoalTown::StopMonitors()
{
	for (local i = 0; i< ::CargoList.len(); i++)
	{
		for (local cid = GSCompany.COMPANY_FIRST; cid <= GSCompany.COMPANY_LAST; cid++)
		{
			if (GSCompany.ResolveCompanyID(cid) != GSCompany.COMPANY_INVALID)
			{
				GSCargoMonitor.GetTownDeliveryAmount(cid, ::CargoList[i],this.id,false);
			}
		}
	}
}
